/* diversity-apdb-backend.c -
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <diversity-world.h>
#include <diversity-bard.h>
#include <diversity-ap.h>
#include <libsoup/soup.h>
#include <string.h>

static struct {
	DiversityWorld *world;
	SoupSession *session;

	const gchar *http_uri;
	const gchar *http_method;

	SoupMessage *query;
	double last_lon, last_lat;
	double query_lon, query_lat;

	DiversityAp *xxx_ap;
} apdb;

typedef struct {
	double lon, lat;
	const gchar *hw_addr;
	const gchar *ssid;
} ApDesc;

/*
 * query: (lat, lon, radius)
 * -> array of (lat, lon, mac, ssid)
 *
 * update: array of (lat, lon, mac, ssid)
 */

gchar *build_method_query(double lon, double lat, double radius)
{
	gchar *body;
	GValue vals[3] = { { 0 } };

	g_value_init(&vals[0], G_TYPE_DOUBLE);
	g_value_init(&vals[1], G_TYPE_DOUBLE);
	g_value_init(&vals[2], G_TYPE_DOUBLE);

	g_value_set_double(&vals[0], lat);
	g_value_set_double(&vals[1], lon);
	g_value_set_double(&vals[2], radius);

	body = soup_xmlrpc_build_method_call("ap.query", vals, 3);

	//g_debug("Query\n%s\n", body);

	return body;
}

gchar *build_method_update(ApDesc *aps, guint num_aps)
{
	gchar *body;
	GValue val = { 0 };
	GValueArray *array;
	guint i;
	
	array = soup_value_array_new();

	for (i = 0; i < num_aps; i++) {
		GValueArray *ap;

		ap = soup_value_array_new_with_vals(
				G_TYPE_DOUBLE, aps[i].lat,
				G_TYPE_DOUBLE, aps[i].lon,
				G_TYPE_STRING, aps[i].hw_addr,
				G_TYPE_STRING, aps[i].ssid,
				G_TYPE_INVALID, NULL);

		soup_value_array_append(array,
				G_TYPE_VALUE_ARRAY, ap);

		g_value_array_free(ap);
	}

	g_value_init(&val, G_TYPE_VALUE_ARRAY);
	g_value_set_static_boxed(&val, array);

	body = soup_xmlrpc_build_method_call("ap.update", &val, 1);

	g_value_array_free(array);

	//g_debug("Update\n%s\n", body);

	return body;
}

static SoupMessage *apdb_message_new(const gchar *body)
{
	SoupMessage *msg;

	if (!body)
		return NULL;

	msg = soup_message_new(apdb.http_method, apdb.http_uri);
	if (!msg)
		return NULL;

	soup_message_set_request(msg, "text/xml", SOUP_MEMORY_TAKE,
				body, strlen(body));

	return msg;
}

static gboolean ap_demarshal(GValueArray *array, ApDesc *ap)
{
	if (array->n_values != 4)
		return FALSE;

	if (G_VALUE_TYPE(array->values + 0) != G_TYPE_DOUBLE ||
	    G_VALUE_TYPE(array->values + 1) != G_TYPE_DOUBLE ||
	    G_VALUE_TYPE(array->values + 2) != G_TYPE_STRING ||
	    G_VALUE_TYPE(array->values + 3) != G_TYPE_STRING)
		return FALSE;

	ap->lat = g_value_get_double(array->values);
	ap->lon = g_value_get_double(array->values + 1);
	ap->hw_addr = g_value_get_string(array->values + 2);
	ap->ssid = g_value_get_string(array->values + 3);

	if (!ap->hw_addr)
		return FALSE;

	return TRUE;
}

static void ap_position(DiversityAp *ap, gdouble lon, gdouble lat)
{
	gdouble len;

	len = DIVERSITY_UNITS(200);
	lon -= len / 2;
	lat -= len / 2;

	diversity_object_geometry_set(DIVERSITY_OBJECT(ap),
			lon, lat, len, len);
}

static gboolean ap_add(GValueArray *array)
{
	DiversityAp *ap;
	ApDesc desc;

	if (!ap_demarshal(array, &desc)) {
		g_debug("failed to demarshal ap");

		return FALSE;
	}

	ap = (DiversityAp *) diversity_world_lookup_object(apdb.world,
			DIVERSITY_OBJECT_TYPE_AP, desc.hw_addr);
	if (!ap) {
		ap = diversity_ap_new();
		DIVERSITY_OBJECT(ap)->id = g_strdup(desc.hw_addr);

		diversity_ap_set_ssid(ap, desc.ssid);
		ap_position(ap, desc.lon, desc.lat);

		diversity_world_add_object(apdb.world, DIVERSITY_OBJECT(ap));

		g_debug("new AP %s added", desc.ssid);
	}

	return TRUE;
}

static gboolean parse_response(SoupMessage *msg, GValue *val, GError **error)
{
	if (!SOUP_STATUS_IS_SUCCESSFUL(msg->status_code)) {
		g_set_error(error, 0, 0, "ERROR %d: %s",
				msg->status_code, msg->reason_phrase);

		return FALSE;
	}

	//g_debug("parse response\n%s\n", msg->response_body->data);

	if (!soup_xmlrpc_parse_method_response(
				msg->response_body->data,
				msg->response_body->length,
				val, error)) {
		if (!*error)
			g_set_error(error, 0, 0, "failed to parse response");

		return FALSE;
	}

	return TRUE;
}

static void on_query_done(SoupSession *session, SoupMessage *msg, gpointer user_data)
{
	GValueArray *array;
	guint i;
	GValue retval = { 0 };
	GError *error = NULL;

	if (!parse_response(msg, &retval, &error)) {
		g_debug("query for (%f, %f) failed: %s",
				apdb.query_lon, apdb.query_lat,
				error->message);

		g_error_free(error);

		return;
	}

	apdb.last_lon = apdb.query_lon;
	apdb.last_lat = apdb.query_lat;

	array = g_value_get_boxed(&retval);

	g_debug("got %d APs", array->n_values);

	for (i = 0; i < array->n_values; i++) {
		GValueArray *ap;

		if (G_VALUE_TYPE(array->values + i) != G_TYPE_VALUE_ARRAY) {
			g_debug("AP %d is of type %s", i,
					G_VALUE_TYPE_NAME(array->values + i));

			continue;
		}

		ap = g_value_get_boxed(array->values + i);
		ap_add(ap);
	}

	g_value_unset(&retval);
}

static void on_update_done(SoupSession *session, SoupMessage *msg, gpointer user_data)
{
	GValue retval = { 0 };
	GError *error = NULL;

	if (!parse_response(msg, &retval, &error)) {
		g_debug("update failed: %s", error->message);
		g_error_free(error);

		return;
	}

	g_debug("AP updated");
}

static void on_geometry_changed(DiversityObject *obj, gdouble lon, gdouble lat, gdouble width, gdouble height)
{
	SoupMessage *msg;
	gchar *body;
	gdouble len;

	if (!diversity_world_get_tunneled(apdb.world)) {
		g_debug("no connection");

		return;
	}

	len = DIVERSITY_UNITS(100);

	if ((lon - apdb.last_lon) * (lon - apdb.last_lon) +
	    (lat - apdb.last_lat) * (lat - apdb.last_lat) < len * len)
		return;

	body = build_method_query(lon, lat, 200.0);
	msg = apdb_message_new(body);
	if (msg) {
		if (apdb.query)
			soup_session_cancel_message(apdb.session,
					apdb.query, SOUP_STATUS_CANCELLED);

		apdb.query_lon = lon;
		apdb.query_lat = lat;

		apdb.query = msg;
		soup_session_queue_message(apdb.session, msg, on_query_done, NULL);
	}
}

static void on_tunneled_changed(DiversityWorld *world, gboolean tunneled)
{
	SoupMessage *msg;
	DiversityObject *obj;
	ApDesc ap;
	gdouble w, h;
	gchar *body;

	if (!tunneled)
		return;

	//obj = diversity_world_get_connected_ap(apdb.world);
	obj = DIVERSITY_OBJECT(apdb.xxx_ap);
	if (!obj) {
		g_debug("no AP connected");

		return;
	}

	diversity_object_geometry_get(obj, &ap.lon, &ap.lat, &w, &h);
	ap.lon += w / 2;
	ap.lat += h / 2;

	ap.hw_addr = diversity_ap_get_hw_address(DIVERSITY_AP(obj));
	ap.ssid = diversity_ap_get_ssid(DIVERSITY_AP(obj));

	body = build_method_update(&ap, 1);
	msg = apdb_message_new(body);
	if (msg)
		soup_session_queue_message(apdb.session, msg, on_update_done, NULL);
}

int apdb_init(DiversityWorld *world, GError **error)
{
	DiversityBard *self;

	if (!g_thread_supported())
		g_thread_init(NULL);

	apdb.world = world;
	g_object_ref(world);

	apdb.session = soup_session_async_new();

	g_signal_connect(world, "tunneled-changed",
			G_CALLBACK(on_tunneled_changed), NULL);

	self = diversity_world_get_self(world);
	g_signal_connect(self, "geometry-changed",
			G_CALLBACK(on_geometry_changed), NULL);

	apdb.http_uri = "http://localhost:12345";
	apdb.http_method = SOUP_METHOD_POST;

	{
		DiversityAp *ap;

		ap = diversity_ap_new();
		diversity_object_geometry_set(DIVERSITY_OBJECT(ap), 1.0, 1.0, 0.0, 0.0);
		diversity_ap_set_hw_address(ap, "01:23:45:67:89:ab");
		diversity_ap_set_ssid(ap, "hahaha");

		diversity_world_add_object(world, DIVERSITY_OBJECT(ap));

		/* test `update' */
		apdb.xxx_ap = ap;
		diversity_world_connect_ap(world, ap);
		diversity_world_set_tunneled(world, TRUE);

		/* test `query' */
		diversity_object_geometry_set(DIVERSITY_OBJECT(self), 1.0, 1.0, 0.0, 0.0);
		diversity_object_geometry_set(DIVERSITY_OBJECT(self), 2.0, 1.0, 0.0, 0.0);

	}

	return TRUE;
}

void apdb_fini(void)
{
	if (apdb.session) {
		g_object_unref(apdb.session);
		apdb.session = NULL;
	}

	if (apdb.world) {
		g_object_unref(G_OBJECT(apdb.world));
		apdb.world = NULL;
	}
}
