/* diversity-iwlib-backend.c -
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <diversity-iwlib-backend.h>
#include <diversity-utils.h>
#include <errno.h>
#include <string.h> /* for memset */
#include "iwlib.h"

struct iwlib_scan {
	gint fd;
	gchar *ifname;

	guint retry_delay;
	guint retry_interval;
	guint retry_id;
	guint retry_count;
	GByteArray *array;
	guint max_size;

	gboolean scanning;
	gboolean has_range;
	struct iw_range range;

	DiversityAp *ap;
};

static struct {
	DiversityWorld *world;
	gint fd;
	guint scan_id;
	GSList *devices;
} iwlib;

static void position_ap(DiversityAp *ap)
{
	gdouble lon, lat, len;

	diversity_world_get_position(iwlib.world, &lon, &lat);

	len = DIVERSITY_UNITS(200);
	lon -= len / 2;
	lat -= len / 2;

	diversity_object_geometry_set(DIVERSITY_OBJECT(ap),
			lon, lat, len, len);
	diversity_object_set_accuracy(DIVERSITY_OBJECT(ap),
			DIVERSITY_OBJECT_ACCURACY_DERIVED);
}

static void scan_commit(struct iwlib_scan *scan)
{
	DiversityAp *ap;
	const gchar *id;

	if (!scan->ap)
		return;

	id = diversity_ap_get_hw_address(scan->ap);
	ap = (DiversityAp *) diversity_world_lookup_object(iwlib.world,
			DIVERSITY_OBJECT_TYPE_AP, id);
	if (ap) {
		g_object_unref(scan->ap);
		scan->ap = NULL;

		return;
	}

	ap = scan->ap;
	scan->ap = NULL;

	DIVERSITY_OBJECT(ap)->id = g_strdup(id);
	position_ap(ap);

	diversity_world_add_object(iwlib.world, DIVERSITY_OBJECT(ap));

	g_debug("commit ap for %s: %c%s (%s)\n",
			scan->ifname,
			diversity_ap_get_flags(ap) ? '*' : ' ',
			diversity_ap_get_ssid(ap),
			id);
}

static void scan_parse_event(struct iwlib_scan *scan, struct iw_event *e)
{
	gchar buf[128];

	/* dangling events */
	if (e->cmd != SIOCGIWAP && !scan->ap)
		return;

	switch (e->cmd) {
	case SIOCGIWAP:
		{
			gchar *hw_addr;

			if (scan->ap)
				scan_commit(scan);

			hw_addr = iw_saether_ntop(&e->u.ap_addr, buf);

			scan->ap = diversity_ap_new();
			diversity_ap_set_hw_address(scan->ap, hw_addr);
		}
		break;
	case SIOCGIWESSID:
		{
			gchar essid[IW_ESSID_MAX_SIZE + 1];
			gint index;

			if (e->u.essid.pointer && e->u.essid.length &&
					e->u.essid.flags) {
				memcpy(essid, e->u.essid.pointer,
						e->u.essid.length);
				essid[e->u.essid.length] = '\0';
			} else {
				*essid = '\0';
			}

			index = e->u.essid.flags & IW_ENCODE_INDEX;

			diversity_ap_set_ssid(scan->ap, essid);
		}
		break;
	case SIOCGIWENCODE:
		{
			guint32 flags;

			flags = (e->u.data.flags & IW_ENCODE_DISABLED)
				? 0
				: DIVERSITY_AP_FLAG_PRIVACY;

			diversity_ap_set_flags(scan->ap, flags);
		}
		break;
	case IWEVQUAL:
		if (scan->has_range &&
		    !(e->u.qual.updated & IW_QUAL_QUAL_INVALID)) {
			gint8 str = 0;

			guint8 updated = e->u.qual.updated;

			if ((updated & (IW_QUAL_DBM | IW_QUAL_RCPI)) ||
			    e->u.qual.level != 0)
				str = e->u.qual.qual * 100 /
					scan->range.max_qual.qual;

			diversity_ap_set_strength(scan->ap, str);
		}
		break;
	case SIOCGIWFREQ:
	case SIOCGIWMODE:
	case SIOCGIWRATE:
	case SIOCGIWMODUL:
	case SIOCGIWNWID:
	case SIOCGIWNAME:
	case IWEVCUSTOM:
	default:
		break;
	}
}

static void scan_parse(struct iwlib_scan *scan, guchar *data, guint len)
{
	struct stream_descr stream;
	struct iw_event iwe;
	gint ret;

	if (!len)
		return;

	iw_init_event_stream(&stream, (gchar *) data, len);
	while (1) {
		ret = iw_extract_event_stream(&stream, &iwe,
				scan->range.we_version_compiled);

		if (ret <= 0)
			break;

		scan_parse_event(scan, &iwe);
	}

	if (scan->ap)
		scan_commit(scan);
}

static gboolean scan_poll(gpointer data)
{
	struct iwlib_scan *scan = data;
	struct iwreq wrq;
	guint len = IW_SCAN_MAX_DATA;

	while (len < scan->max_size) {
		g_byte_array_set_size(scan->array, len);

		wrq.u.data.pointer = scan->array->data;
		wrq.u.data.length = len;
		wrq.u.data.flags = 0;

		if (iw_get_ext(scan->fd, scan->ifname, SIOCGIWSCAN, &wrq) >= 0)
			break;

		if (errno == EAGAIN) {
			if (scan->retry_count) {
				scan->retry_count++;

				return TRUE;
			}

			g_source_remove(scan->retry_id);
			scan->retry_count++;

			scan->retry_id =
				g_timeout_add(scan->retry_interval,
						scan_poll, scan);

			return FALSE;
		} else if (errno == E2BIG &&
			   scan->range.we_version_compiled > 16) {
			if (wrq.u.data.length > len)
				len = wrq.u.data.length;
			else
				len *= 2;
		} else {
			return FALSE;
		}
	}

	if (len > scan->max_size)
		return FALSE;

	scan_parse(scan, scan->array->data, wrq.u.data.length);

	scan->scanning = FALSE;
	scan->retry_id = 0;

	return FALSE;
}

static void scan_cancel(struct iwlib_scan *scan)
{
	scan->scanning = FALSE;

	if (scan->retry_id) {
		g_source_remove(scan->retry_id);
		scan->retry_id = 0;
	}
}

static gboolean scan_do(struct iwlib_scan *scan)
{
	struct iwreq wrq;
	
	if (scan->scanning)
		scan_cancel(scan);

	scan->scanning = TRUE;

	if (iw_get_range_info(scan->fd, scan->ifname, &scan->range) >= 0)
		scan->has_range = TRUE;
	else
		scan->has_range = FALSE;

	if (!scan->has_range || scan->range.we_version_compiled < 14)
		return FALSE;

	wrq.u.data.pointer = NULL;
	wrq.u.data.flags = 0;
	wrq.u.data.length = 0;

	if (iw_set_ext(scan->fd, scan->ifname, SIOCSIWSCAN, &wrq) >= 0) {
		g_timeout_add(scan->retry_delay, scan_poll, scan);
	} else {
		if (errno != EPERM)
			return FALSE;

		/* poll the result immediately */
		scan_poll(scan);
	}

	return TRUE;
}

static struct iwlib_scan *scan_new(gint fd, const gchar *ifname)
{
	struct iwlib_scan *scan;

	scan = g_new0(struct iwlib_scan, 1);

	scan->fd = fd;
	scan->ifname = g_strdup(ifname);

	scan->retry_delay = 250;
	scan->retry_interval = 100;

	scan->array = g_byte_array_new();
	scan->max_size = 0x100000;

	return scan;
}

static void scan_destroy(struct iwlib_scan *scan)
{
	scan_cancel(scan);

	g_free(scan->ifname);
	g_byte_array_free(scan->array, TRUE);

	g_free(scan);
}

static gboolean iwlib_scan(gpointer data)
{
	GSList *tmp_list = iwlib.devices;

	while (tmp_list) {
		struct iwlib_scan *scan = tmp_list->data;

		tmp_list = tmp_list->next;

		if (!scan_do(scan)) {
			g_debug("stop scanning %s: scan failed", scan->ifname);

			scan_destroy(scan);
			iwlib.devices = g_slist_remove(iwlib.devices, scan);
		}
	}

	return TRUE;
}

static void iwlib_enum_devices(void)
{
	struct iwlib_scan *scan;

	scan = scan_new(iwlib.fd, "eth0");

	iwlib.devices = g_slist_prepend(iwlib.devices, scan);

	scan = scan_new(iwlib.fd, "wlan0");

	iwlib.devices = g_slist_prepend(iwlib.devices, scan);
}

static void iwlib_free_devices(void)
{
	while (iwlib.devices) {
		scan_destroy(iwlib.devices->data);
		iwlib.devices =
			g_slist_remove_link(iwlib.devices, iwlib.devices);
	}
}

static void on_accuracy_changed(DiversityBard *self)
{
	DiversityObjectAccuracy accuracy;

	accuracy = diversity_object_get_accuracy(DIVERSITY_OBJECT(self));

	if (accuracy == DIVERSITY_OBJECT_ACCURACY_NONE) {
		if (iwlib.scan_id) {
			g_source_remove(iwlib.scan_id);
			iwlib.scan_id = 0;
		}
	} else {
		if (!iwlib.scan_id) {
			iwlib.scan_id =
				g_timeout_add_seconds(15, iwlib_scan, NULL);
			iwlib_scan(NULL);
		}
	}
}

int iwlib_init(DiversityWorld *world, GError **error)
{
	DiversityBard *self;

	iwlib.world = world;
	g_object_ref(world);

	iwlib.fd = iw_sockets_open();
	if (iwlib.fd < 0) {
		g_set_error(error, 0, 0, "failed to open socket");

		return FALSE;
	}

	iwlib_enum_devices();

	self = diversity_world_get_self(iwlib.world);
	g_signal_connect(self, "notify::accuracy",
		G_CALLBACK(on_accuracy_changed), NULL);

	on_accuracy_changed(self);

	return TRUE;
}

void iwlib_fini(void)
{
	iwlib_free_devices();

	if (iwlib.scan_id) {
		g_source_remove(iwlib.scan_id);
		iwlib.scan_id = 0;
	}

	if (iwlib.fd >= 0) {
		iw_sockets_close(iwlib.fd);
		iwlib.fd = -1;
	}
	if (iwlib.world) {
		g_object_unref(G_OBJECT(iwlib.world));
		iwlib.world = NULL;
	}
}
