/* diversity-qtopia-backend.c -
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <dbus/dbus-glib.h>
#include <diversity-world.h>
#include <diversity-bard.h>
#include <diversity-tag.h>

#include "diversity-qtopia-backend.h"
#include "diversity-qtopia-contacts.h"
#include "diversity-qtopia-sms.h"

static struct {
	DiversityWorld *world;
	DiversityQtopiaContacts *contacts;
	DiversityQtopiaSms *sms;
} qtopia;

static void bard_update(DiversityBard *bard, DiversityQtopiaContact *c)
{
	diversity_bard_set_fullname(bard, c->fullname);
	diversity_bard_set_phone(bard, c->phone);

	g_debug("%s updated: %s (%s)\n",
			c->uid, c->fullname, c->phone);
}

static DiversityBard *bard_new_from_contact(DiversityQtopiaContact *c)
{
	DiversityBard *bard;

	bard = diversity_bard_new();

	DIVERSITY_OBJECT(bard)->id = g_strdup(c->uid);
	diversity_bard_set_uid(bard, c->uid);

	bard_update(bard, c);

	return bard;
}

static void add_one(DiversityQtopiaContact *c)
{
	DiversityBard *self, *bard;

	bard = (DiversityBard *) diversity_world_lookup_object(qtopia.world,
			DIVERSITY_OBJECT_TYPE_BARD, c->uid);
	if (bard) {
		bard_update(bard, c);

		return;
	}

	bard = bard_new_from_contact(c);
	diversity_world_add_object(qtopia.world, DIVERSITY_OBJECT(bard));

	self = diversity_world_get_self(qtopia.world);
	diversity_object_connect(DIVERSITY_OBJECT(self),
				 DIVERSITY_OBJECT(bard));
}

static void update_one(DiversityQtopiaContact *c)
{
	DiversityBard *bard;

	bard = (DiversityBard *) diversity_world_lookup_object(qtopia.world,
			DIVERSITY_OBJECT_TYPE_BARD, c->uid);
	if (!bard) {
		g_warning("update non-existing uid %s", c->uid);

		add_one(c);

		return;
	}

	bard_update(bard, c);
}

static void remove_one(const gchar *uid)
{
	DiversityBard *self, *bard;

	bard = (DiversityBard *) diversity_world_lookup_object(qtopia.world,
			DIVERSITY_OBJECT_TYPE_BARD, uid);
	if (!bard)
		return;

	self = diversity_world_get_self(qtopia.world);
	diversity_object_disconnect(DIVERSITY_OBJECT(self),
				    DIVERSITY_OBJECT(bard));

	diversity_world_remove_object(qtopia.world, DIVERSITY_OBJECT(bard));
}

static void on_contacts_added(DiversityQtopiaContacts *qcontacts, GList *contacts, gpointer data)
{
	while (contacts) {
		add_one(contacts->data);
		contacts = contacts->next;
	}
}

static void on_contacts_changed(DiversityQtopiaContacts *qcontacts, GList *contacts, gpointer data)
{
	while (contacts) {
		update_one(contacts->data);
		contacts = contacts->next;
	}
}

static void on_contacts_removed(DiversityQtopiaContacts *qcontacts, GList *uids, gpointer data)
{
	while (uids) {
		remove_one(uids->data);
		uids = uids->next;
	}
}

static void on_tag_added(DiversityQtopiaSms *qsms, const gchar *uid, gdouble lon, gdouble lat, const gchar *desc)
{
	DiversityObject *tag;

	tag = diversity_world_lookup_object(qtopia.world,
			DIVERSITY_OBJECT_TYPE_TAG, uid);
	if (!tag) {
		tag = (DiversityObject *) diversity_tag_new();

		tag->id = g_strdup(uid);
		diversity_object_set_accuracy(tag,
				DIVERSITY_OBJECT_ACCURACY_EXACT);
		diversity_object_geometry_set(tag, lon, lat, 0.0, 0.0);

		diversity_tag_set(DIVERSITY_TAG(tag),
				DIVERSITY_TAG_DESCRIPTION, desc);

		diversity_object_touch(tag);

		diversity_world_add_object(qtopia.world, tag);
	}
}

static void on_object_removed(DiversityWorld *world, DiversityObject *obj)
{
	if (obj->type != DIVERSITY_OBJECT_TYPE_TAG)
		return;

	if (obj->id)
		diversity_qtopia_sms_remove(qtopia.sms, obj->id);
}

int qtopia_init(DiversityWorld *world, GError **error)
{
	DiversityBard *self;

	if (!g_thread_supported()) {
		g_thread_init(NULL);
		dbus_g_thread_init();
	}

	qtopia.world = world;
	g_object_ref(world);

	g_signal_connect(qtopia.world, "object-removed", G_CALLBACK(on_object_removed), NULL);

	qtopia.contacts = diversity_qtopia_contacts_new();
	if (!qtopia.contacts) {
		qtopia_fini();

		return FALSE;
	}

	g_signal_connect(G_OBJECT(qtopia.contacts), "contacts-added",
			 G_CALLBACK(on_contacts_added), NULL);
	g_signal_connect(G_OBJECT(qtopia.contacts), "contacts-removed",
			 G_CALLBACK(on_contacts_removed), NULL);
	g_signal_connect(G_OBJECT(qtopia.contacts), "contacts-changed",
			 G_CALLBACK(on_contacts_changed), NULL);

	if (!diversity_qtopia_contacts_open_and_start(qtopia.contacts)) {
		g_set_error(error, 0, 0, "failed to open contacts");
		qtopia_fini();

		return FALSE;
	}

	qtopia.sms = diversity_qtopia_sms_new("qtopia");
	if (!qtopia.sms) {
		qtopia_fini();

		return FALSE;
	}

	g_signal_connect(qtopia.sms, "tag-added", G_CALLBACK(on_tag_added), NULL);

	if (!diversity_qtopia_sms_open_and_start(qtopia.sms)) {
		g_set_error(error, 0, 0, "failed to open sms");
		qtopia_fini();

		return FALSE;
	}

	self = diversity_world_get_self(qtopia.world);
	diversity_bard_add_equipment(self, DIVERSITY_EQUIPMENT(qtopia.sms));

	return TRUE;
}

void qtopia_fini(void)
{
	if (qtopia.world) {
		g_object_unref(G_OBJECT(qtopia.world));
		qtopia.world = NULL;
	}

	if (qtopia.sms) {
		g_object_unref(qtopia.sms);
		qtopia.sms = NULL;
	}

	if (qtopia.contacts) {
		g_object_unref(qtopia.contacts);
		qtopia.contacts = NULL;
	}
}
