/* diversity-qtopia-contacts.c - DiversityQtopiaContacts
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <dbus/dbus-glib.h>
#include <string.h> /* for strcmp */
#include "diversity-qtopia-contacts.h"
#include "diversity-qtopia.h"

enum {
	CONTACTS_ADDED,
	CONTACTS_CHANGED,
	CONTACTS_REMOVED,
	LAST_SIGNAL
};

#define DIVERSITY_QTOPIA_CONTACTS_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), DIVERSITY_TYPE_QTOPIA_CONTACTS, DiversityQtopiaContactsPrivate))
typedef struct _DiversityQtopiaContactsPrivate DiversityQtopiaContactsPrivate;

struct _DiversityQtopiaContactsPrivate {
	DBusGConnection *connection;
	DBusGProxy *proxy;

	guint open_interval;
	guint open_max_retry;

	guint open_id;
	guint open_retry;
	gboolean opened;

	GList *contacts;
};

static guint contacts_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(DiversityQtopiaContacts, diversity_qtopia_contacts, G_TYPE_OBJECT);

static void contact_destroy(DiversityQtopiaContact *c);

static void diversity_qtopia_contacts_dispose(GObject *obj)
{
	DiversityQtopiaContactsPrivate *priv =
		DIVERSITY_QTOPIA_CONTACTS_GET_PRIVATE(obj);

	if (priv->open_id) {
		g_source_remove(priv->open_id);
		priv->open_id = 0;
	}

	if (priv->proxy) {
		if (priv->opened) {
			dbus_g_proxy_call(priv->proxy,
					  "close", NULL,
					  G_TYPE_INVALID,
					  G_TYPE_INVALID);

			priv->opened = FALSE;
		}

		g_object_unref(priv->proxy);
		priv->proxy = NULL;
	}

	((GObjectClass *) diversity_qtopia_contacts_parent_class)->dispose(obj);
}

static void diversity_qtopia_contacts_finalize(GObject *obj)
{
	DiversityQtopiaContactsPrivate *priv =
		DIVERSITY_QTOPIA_CONTACTS_GET_PRIVATE(obj);

	if (priv->connection)
		dbus_g_connection_unref(priv->connection);

	g_list_foreach(priv->contacts, (GFunc) contact_destroy, NULL);
	g_list_free(priv->contacts);

	((GObjectClass *) diversity_qtopia_contacts_parent_class)->finalize(obj);
}

static void diversity_qtopia_contacts_class_init(DiversityQtopiaContactsClass *klass)
{
	GObjectClass *o_class = (GObjectClass *) klass;

	o_class->dispose = diversity_qtopia_contacts_dispose;
	o_class->finalize = diversity_qtopia_contacts_finalize;

	g_type_class_add_private(klass, sizeof(DiversityQtopiaContactsPrivate));

	contacts_signals[CONTACTS_ADDED] =
		g_signal_new("contacts-added",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityQtopiaContactsClass, contacts_added),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__POINTER,
			     G_TYPE_NONE, 1,
			     G_TYPE_POINTER);

	contacts_signals[CONTACTS_CHANGED] =
		g_signal_new("contacts-changed",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityQtopiaContactsClass, contacts_changed),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__POINTER,
			     G_TYPE_NONE, 1,
			     G_TYPE_POINTER);

	contacts_signals[CONTACTS_REMOVED] =
		g_signal_new("contacts-removed",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityQtopiaContactsClass, contacts_removed),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__POINTER,
			     G_TYPE_NONE, 1,
			     G_TYPE_POINTER);
}

static void diversity_qtopia_contacts_init(DiversityQtopiaContacts *qcontacts)
{
	DiversityQtopiaContactsPrivate *priv =
		DIVERSITY_QTOPIA_CONTACTS_GET_PRIVATE(qcontacts);

	priv->connection = NULL;
	priv->proxy = NULL;

	priv->open_interval = 5;
	priv->open_max_retry = 12;

	priv->open_id = 0;
	priv->open_retry = 0;
	priv->opened = FALSE;

	priv->contacts = NULL;
}

static DiversityQtopiaContact *contact_new(const gchar *uid)
{
	DiversityQtopiaContact *c;

	c = g_slice_new0(DiversityQtopiaContact);
	c->uid = g_strdup(uid);

	return c;
}

static void contact_destroy(DiversityQtopiaContact *c)
{
	g_free(c->uid);
	g_free(c->fullname);
	g_free(c->phone);
	g_slice_free(DiversityQtopiaContact, c);
}

static gint contact_compare(gconstpointer a, gconstpointer b)
{
	const DiversityQtopiaContact *c1, *c2;

	c1 = a;
	c2 = b;

	return strcmp(c1->uid, c2->uid);
}

static gboolean contact_query(DiversityQtopiaContact *c, DBusGProxy *proxy)
{
	GError *error = NULL;
	GType out_type;
	GHashTable *dict;
	const gchar *phone_keys[] = {
		"homeMobile",
		"businessMobile",
		"homePhone",
		"businessPhone",
		NULL
	};
	const gchar **p;
	const gchar *phone;
	GValue *val;

	out_type = dbus_g_type_get_map(
			"GHashTable", G_TYPE_STRING, G_TYPE_VALUE);

	if (!dbus_g_proxy_call(proxy,
			       "contact", &error,
			       G_TYPE_STRING, c->uid,
			       G_TYPE_INVALID,
			       out_type, &dict,
			       G_TYPE_INVALID)) {
		g_warning("failed to query contact %s: %s",
				c->uid, error->message);

		g_error_free(error);

		return FALSE;
	}

	val = g_hash_table_lookup(dict, "firstName");
	c->fullname = (gchar *) g_value_get_string(val);
	val = g_hash_table_lookup(dict, "lastName");
	c->fullname = g_strconcat(
			c->fullname,
			" ",
			g_value_get_string(val),
			NULL);

	p = phone_keys;
	while (*p) {
		val = g_hash_table_lookup(dict, *p);
		phone = g_value_get_string(val);
		if (phone && *phone) {
			c->phone = g_strdup(phone);

			break;
		}

		p++;
	}

	g_hash_table_destroy(dict);

	return TRUE;
}

DiversityQtopiaContacts *diversity_qtopia_contacts_new(void)
{
	return g_object_new(DIVERSITY_TYPE_QTOPIA_CONTACTS, NULL);
}

static GList *diversity_qtopia_contacts_list(DiversityQtopiaContacts *qcontacts)
{
	DiversityQtopiaContactsPrivate *priv =
		DIVERSITY_QTOPIA_CONTACTS_GET_PRIVATE(qcontacts);
	GPtrArray *array;
	GList *contacts = NULL;
	guint i;
	GError *error = NULL;
	GType out_type;

	out_type = dbus_g_type_get_collection("GPtrArray", G_TYPE_VALUE);

	if (!dbus_g_proxy_call(priv->proxy,
				"listContacts", &error,
				G_TYPE_INVALID,
				out_type, &array,
				G_TYPE_INVALID)) {
		g_warning("failed to list contacts: %s", error->message);

		g_error_free(error);

		return NULL;
	}

	for (i = 0; i < array->len; i++) {
		GValue *val = g_ptr_array_index(array, i);
		const gchar *uid;
		DiversityQtopiaContact *c;

		uid = g_value_get_string(val);
		c = contact_new(uid);

		if (contact_query(c, priv->proxy))
			contacts = g_list_prepend(contacts, c);
		else
			contact_destroy(c);

		g_value_unset(val);
	}

	g_ptr_array_free(array, TRUE);

	return contacts;
}

static void on_model_changed(DBusGProxy *proxy, DiversityQtopiaContacts *qcontacts)
{
	DiversityQtopiaContactsPrivate *priv =
		DIVERSITY_QTOPIA_CONTACTS_GET_PRIVATE(qcontacts);
	GList *contacts;
	GList *added, *removed, *changed, *tmp_list;

	contacts = diversity_qtopia_contacts_list(qcontacts);

	added = contacts;
	removed = changed = NULL;

	tmp_list = priv->contacts;
	while (tmp_list) {
		DiversityQtopiaContact *c = tmp_list->data;
		GList *tmp;

		tmp = g_list_find_custom(added, c, contact_compare);
		if (tmp) {
			added = g_list_remove_link(added, tmp);
			changed = g_list_concat(tmp, changed);
		} else {
			/* note that UIDs are owned by priv->contacts */
			removed = g_list_prepend(removed, c->uid);
		}

		tmp_list = tmp_list->next;
	}

	g_debug("model changed: %d removed, %d changed, %d added",
			g_list_length(removed),
			g_list_length(changed),
			g_list_length(added));

	if (removed)
		g_signal_emit(qcontacts,
			      contacts_signals[CONTACTS_REMOVED], 0,
			      removed);

	if (changed)
		g_signal_emit(qcontacts,
			      contacts_signals[CONTACTS_CHANGED], 0,
			      changed);

	if (added)
		g_signal_emit(qcontacts,
			      contacts_signals[CONTACTS_ADDED], 0,
			      added);

	g_list_foreach(priv->contacts, (GFunc) contact_destroy, NULL);
	g_list_free(priv->contacts);

	priv->contacts = g_list_concat(added, changed);
}

static gboolean timeout_open(gpointer data)
{
	DiversityQtopiaContacts *qcontacts = data;
	DiversityQtopiaContactsPrivate *priv =
		DIVERSITY_QTOPIA_CONTACTS_GET_PRIVATE(qcontacts);

	if (priv->opened) {
		priv->open_id = 0;

		return FALSE;
	}

	if (dbus_g_proxy_call(priv->proxy,
			      "open", NULL,
			      G_TYPE_INVALID,
			      G_TYPE_INVALID)) {
		g_debug("contacts opened");

		dbus_g_proxy_add_signal(priv->proxy,
					"modelChanged",
					G_TYPE_INVALID);
		dbus_g_proxy_connect_signal(priv->proxy,
					    "modelChanged",
					    G_CALLBACK(on_model_changed),
					    qcontacts, NULL);

		on_model_changed(priv->proxy, qcontacts);

		priv->opened = TRUE;
		priv->open_id = 0;

		return FALSE;
	} else {
		priv->open_retry++;

		if (priv->open_retry >= priv->open_max_retry) {
			g_warning("failed to open contacts");

			priv->open_id = 0;

			return FALSE;
		}
	}

	return TRUE;
}

gboolean diversity_qtopia_contacts_open_and_start(DiversityQtopiaContacts *qcontacts)
{
	DiversityQtopiaContactsPrivate *priv;
	GError *error = NULL;
	
	priv = DIVERSITY_QTOPIA_CONTACTS_GET_PRIVATE(qcontacts);
	if (priv->opened || priv->open_id)
		return TRUE;

	priv->connection = dbus_g_bus_get(QTOPIA_DBUS_BUS, &error);
	if (!priv->connection) {
		g_warning("failed to get bus: %s", error->message);

		g_error_free(error);

		return FALSE;
	}

	priv->proxy =
		dbus_g_proxy_new_for_name(priv->connection,
					  QTOPIA_DBUS_SERVICE,
					  QTOPIA_CONTACTS_DBUS_PATH,
					  QTOPIA_CONTACTS_DBUS_INTERFACE);

	if (!priv->proxy) {
		g_warning("failed to get proxy");

		dbus_g_connection_unref(priv->connection);
		priv->connection = NULL;

		return FALSE;
	}

	if (timeout_open(qcontacts))
		priv->open_id = g_timeout_add_seconds(priv->open_interval,
				timeout_open, qcontacts);

	return TRUE;
}
